"""
Acoustics
=========

The acoustics module.

"""
import acoustics.aio
import acoustics.ambisonics
import acoustics.atmosphere
import acoustics.bands
import acoustics.building
import acoustics.cepstrum
import acoustics.criterion
import acoustics.decibel
import acoustics.descriptors
import acoustics.directivity
import acoustics.doppler
import acoustics.generator
import acoustics.imaging
import acoustics.octave
import acoustics.power
import acoustics.quantity
import acoustics.reflection
import acoustics.room
import acoustics.signal
#import acoustics.utils
import acoustics.weighting

from acoustics._signal import Signal
